/*
* Copyright (c) 2005-08, Craig Bishop and Colin Graeme Johnson
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the names of Craig Bishop and/or Colin Graeme Johnson, 
*       nor the name of the University of Kent may be used to endorse 
*       or promote products derived from this software without
*       specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY COLIN GRAEME JOHNSON ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL COLIN GRAEME JOHNSON BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
 * Created on 06-Jul-2005
 *
 */
package main.rules;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Set;

import main.progAnal.ConditionChecker;

/**
 * @author cbishop
 */
public abstract class RoleChecker {

	protected String variable;

	protected LinkedHashMap analysedMap;

	protected ArrayList methods;

	protected ConditionChecker conditionChecker;

	protected ResultStringer resultStringer;

	//condition storage
	protected HashMap assignmentsInLoop;

	protected HashMap usagesInLoop;

	protected HashMap conditionalsInLoop;

	protected HashMap usageInAssignmentLoop;

	protected HashMap conditionalInAssignmentLoop;

	protected HashMap assignmentsInBranch;

	protected HashMap conditionalAssignmentBranch;

	protected ArrayList onBothSides;

	protected ArrayList toggleStatements;

	protected ArrayList incDecStatements;

	protected HashMap useOutsideAssign;

	protected ArrayList assignBeforeUse;

	protected HashMap condOutsideAssign;

	protected ArrayList indirectBothSides;

	protected ArrayList reorganize;

	protected ArrayList transformations;

	protected ArrayList nestedBooleanAssignment;

	protected ArrayList methodAssignments;

	protected ArrayList otherAssignments;

	protected ArrayList assignmentInFor;

	protected ArrayList useForAssignLoopCondition;

	protected ArrayList isArray;

	//condition flags
	protected boolean onlyOneAssignment;

	protected boolean isAssignedInFor;

	protected boolean assignedInLoop;

	protected boolean usedInAssignLoop;

	protected boolean conditionalUseInAssign;

	protected boolean indirectUse;

	protected boolean noDirectUsage;

	protected boolean conditionForAssignBranch;

	protected boolean onBothSidesOfAssign;

	protected boolean toggleStatement;

	protected boolean incDecStatement;

	protected boolean usedOutsideAssign;

	protected boolean assignedBeforeUse;

	protected boolean loopUseNoAssign;

	protected boolean assignCondLoop;

	protected boolean assignmentLoopConditionUse;

	protected boolean indirectBS;

	protected boolean isReorganize;

	protected boolean isArrayVariable;

	protected boolean isTransform;

	protected boolean nestedBooleanAssign;

	protected boolean methodAssignment;

	protected boolean otherAssign;

	//result Flags
	protected boolean isFixedValue;

	protected boolean isOrganizer;

	protected boolean isStepper;

	protected boolean isMostRecentHolder;

	protected boolean isGatherer;

	protected boolean isMostWantedHolder;

	protected boolean isOneWayFlag;

	protected boolean isTransformation;

	protected boolean isFollower;

	protected boolean isTemporary;

	//Explanatory strings
	private String offendingStatement;

	private String reason;

	/**
	 * Constructor for RoleChecker
	 * 
	 * @param inputMap
	 *            LinkedHashMap containing analysed statements
	 * @param var
	 *            String being variable for which role is to be checked
	 * @param meths
	 *            ArrayList of method names in source code
	 */
	public RoleChecker(LinkedHashMap inputMap, String var, ArrayList meths) {
		variable = var;
		analysedMap = inputMap;
		methods = meths;
		conditionChecker = new ConditionChecker(analysedMap, variable, methods);
		resultStringer = new ResultStringer(variable);
		initialiseConditionFlags();
		getConditions();
		setConditionFlags();
	}

	/**
	 * Abstract method to be completed in each Class extending RoleChecker
	 * 
	 * @return ArrayList of results from role checking
	 */
	public abstract ArrayList checkRole();

	/*
	 * Initialise condition flags
	 */
	protected void initialiseConditionFlags() {
		//variable assigned in loop
		assignedInLoop = false;
		//variable used in assignment loop
		usedInAssignLoop = false;
		//variable used conditionally in assignment loop
		conditionalUseInAssign = false;
		//variable used indirectly for assignment loop condition
		indirectUse = false;
		//variable only assigned in "for" loop statement
		isAssignedInFor = false;
		//no direct usage of variable
		noDirectUsage = false;
		//assignment in branch for which variable is condition
		conditionForAssignBranch = false;
		//variable appears on both sides of assignment statement
		onBothSidesOfAssign = false;
		//variable toggled within loop
		toggleStatement = false;
		//variable incremented etc within loop
		incDecStatement = false;
		//variable is not used outside of loop in which it is assigned
		usedOutsideAssign = false;
		//variable is not assigned before it is used in assignment loop
		assignedBeforeUse = false;
		//variable is used as condition for assignment loop
		assignCondLoop = false;
		//variable used conditionally for loop outside of assignment loop
		loopUseNoAssign = false;
		//variable not used indirectly on both sides of assignment statement
		indirectBS = false;
		//variable is reorganize array
		isReorganize = false;
		//variable is Array
		isArrayVariable = false;
		//variable is not used in transformation statement
		isTransform = false;
		//variable is not assigned twice in loop, one assignment being in
		// nested loop
		nestedBooleanAssign = false;
		//variable is not assigned with output from method
		methodAssignment = false;
		otherAssign = false;
	}

	/*
	 * Get conditions under which variables are assigned and used in program
	 */
	protected void getConditions() {
		onlyOneAssignment = conditionChecker.onlyOneAssignmentStatement();
		//find assignments in loop
		assignmentsInLoop = conditionChecker.foundIn("assignment", "loop");
		//find direct usage in loop
		usagesInLoop = conditionChecker.foundIn("usage", "loop");
		//find conditional usage in loop
		conditionalsInLoop = conditionChecker.foundIn("conditional", "loop");
		//find direct usage in assignment loop
		usageInAssignmentLoop = conditionChecker.foundInAssignmentLoop(
				assignmentsInLoop, usagesInLoop);
		//find conditional usage in assignment loop
		conditionalInAssignmentLoop = conditionChecker.foundInAssignmentLoop(
				assignmentsInLoop, conditionalsInLoop);
		//find assignment in branch
		assignmentsInBranch = conditionChecker.foundIn("assignment", "branch");
		//find incidents where assignment is dependent on condition of variable
		conditionalAssignmentBranch = conditionChecker
				.conditionForAssignmentBranch(assignmentsInBranch,
						conditionalsInLoop);
		//if not incidents where assignment dependent on condition of variable
		// check assignment statement for leading branch condition
		if (conditionalAssignmentBranch.size() < 1)
			conditionalAssignmentBranch = conditionChecker
					.conditionAssignStatement();
		//find assignment statements where variable appears on both sides of
		// "="
		onBothSides = conditionChecker.statementTypeCheck((HashMap) analysedMap
				.get("assignment"), "bothSides");
		//find assignment statements in loop where variable is toggled
		toggleStatements = conditionChecker.statementTypeCheck(
				assignmentsInLoop, "toggle");
		//find assignment statements in loop where variable is incremented,
		// decremented, etc
		incDecStatements = conditionChecker.statementTypeCheck(
				assignmentsInLoop, "incDec");
		//find use outside of assignment loop
		useOutsideAssign = conditionChecker.outsideAssignLoop(
				usageInAssignmentLoop, "usage");
		//find whether variable used before assigned in loop
		assignBeforeUse = conditionChecker
				.assignBeforeUse(usageInAssignmentLoop);
		//find conditional use for loop statements loops outside of assignment
		// loop
		condOutsideAssign = conditionChecker.outsideAssignLoop(
				conditionalInAssignmentLoop, "conditional");
		//check if variable only assigned value as part of "for" loop statement
		assignmentInFor = conditionChecker.assignmentInFor("usage");
		//check if variable only assigned value as part of "for" loop statement
		if (assignmentInFor.size() < 1)
			assignmentInFor = conditionChecker.assignmentInFor("conditional");
		//find assignment statements in loop where variable appears indirectly
		// on both sides of "=")
		indirectBothSides = conditionChecker.statementTypeCheck(
				assignmentsInLoop, "indirect");
		reorganize = conditionChecker.reorganize();
		//check if variable is array
		isArray = conditionChecker.arrayCheck();
		//check if variable is only assigned by transformation statement
		transformations = conditionChecker.transform(assignmentsInLoop);
		nestedBooleanAssignment = conditionChecker
				.nestedBooleanAssign(assignmentsInLoop);
		//check is variable is assigned with output from method
		methodAssignments = conditionChecker
				.assignedWithMethod(assignmentsInLoop);
		//check if variable is assigned with value other than of other variable
		otherAssignments = conditionChecker.otherAssignment(assignmentsInLoop);
	}

	/*
	 * Set condition flags depending on conditions
	 */
	protected void setConditionFlags() {
		if (assignmentsInLoop.size() > 0)
			assignedInLoop = true;
		if (usageInAssignmentLoop.size() > 0)
			usedInAssignLoop = true;
		if (conditionalInAssignmentLoop.size() > 0)
			conditionalUseInAssign = true;
		if (conditionalAssignmentBranch.size() > 0)
			conditionForAssignBranch = true;
		if (onBothSides.size() > 0)
			onBothSidesOfAssign = true;
		if (toggleStatements.size() > 0)
			toggleStatement = true;
		if (incDecStatements.size() > 0)
			incDecStatement = true;
		if (conditionalUseInAssign) {
			useForAssignLoopCondition = conditionChecker
					.useForLoopCondition(conditionalInAssignmentLoop);
			assignCondLoop = false;
			if (useForAssignLoopCondition.size() > 0)
				assignCondLoop = true;
			if (!assignCondLoop) {
				indirectUse = conditionChecker
						.indirectUse(conditionalInAssignmentLoop);
			}
		}
		if (assignmentInFor.size() > 0)
			isAssignedInFor = true;
		if (conditionChecker.noDirectUsage() && !onBothSidesOfAssign)
			noDirectUsage = true;
		if (useOutsideAssign.size() > 0)
			usedOutsideAssign = true;
		if (assignBeforeUse.size() > 0)
			assignedBeforeUse = true;
		if (condOutsideAssign.size() > 0)
			loopUseNoAssign = true;
		assignmentLoopConditionUse = (indirectUse || assignCondLoop);
		if (indirectBothSides.size() > 0)
			indirectBS = true;
		if (reorganize.size() > 0)
			isReorganize = true;
		if (transformations.size() > 0)
			isTransform = true;
		if (nestedBooleanAssignment.size() > 0)
			nestedBooleanAssign = true;
		if (methodAssignments.size() > 0)
			methodAssignment = true;
		if (otherAssignments.size() > 0)
			otherAssign = true;
		if (isArray.size() > 0)
			isArrayVariable = true;
	}

	/*
	 * Return if variable is fixed value
	 */
	protected boolean isFixedValue() {
		fixedValue();
		return isFixedValue;
	}

	/*
	 * Return if variable is organizer
	 */
	protected boolean isOrganizer() {
		organizer();
		return isOrganizer;
	}

	/*
	 * Return if variable is stepper
	 */
	protected boolean isStepper() {
		stepper();
		return isStepper;
	}

	/*
	 * Return if variable is most recent holder
	 */
	protected boolean isMostRecentHolder() {
		mostRecentHolder();
		return isMostRecentHolder;
	}

	/*
	 * Return if variable is gatherer
	 */
	protected boolean isGatherer() {
		gatherer();
		return isGatherer;
	}

	/*
	 * Return if variable is most wanted holder
	 */
	protected boolean isMostWantedHolder() {
		mostWantedHolder();
		return isMostWantedHolder;
	}

	/*
	 * Return if variable is one way flag
	 */
	protected boolean isOneWayFlag() {
		oneWayFlag();
		return isOneWayFlag;
	}

	/*
	 * Return if variable is transformation
	 */
	protected boolean isTransformation() {
		transformation();
		return isTransformation;
	}

	/*
	 * Return if variable is follower
	 */
	protected boolean isFollower() {
		follower();
		return isFollower;
	}

	/*
	 * Return if variable is temporary
	 */
	protected boolean isTemporary() {
		temporary();
		return isTemporary;
	}

	/*
	 * Test whether variable is fixed value
	 */
	protected ArrayList fixedValue() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isFixedValue = true;
		if (isArrayVariable && isReorganize) {
			isFixedValue = false;
			setMessage(1);
		} else if (incDecStatement) {
			isFixedValue = false;
			setMessage(2);
		} else if (toggleStatement) {
			isFixedValue = false;
			setMessage(3);
		} else if (onBothSidesOfAssign) {
			isFixedValue = false;
			setMessage(4);
		} else if (indirectBS) {
			isFixedValue = false;
			setMessage(5);
		} else if (isTransform && usedInAssignLoop) {
			isFixedValue = false;
			setMessage(6);
		} else if (isAssignedInFor) {
			isFixedValue = false;
			setMessage(7);
		} else if (nestedBooleanAssign) {
			isFixedValue = false;
			setMessage(8);
		} else if (conditionForAssignBranch) {
			isFixedValue = false;
			setMessage(9);
		} else if (!methodAssignment && !otherAssign && usedInAssignLoop) {
			isFixedValue = false;
			setMessage(10);
		} else if (usedInAssignLoop && !isArrayVariable) {
			isFixedValue = false;
			Set keySet = usageInAssignmentLoop.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "used in loop in which assigned";
		} else if (conditionalUseInAssign && !assignmentLoopConditionUse
				&& !isArrayVariable) {
			isFixedValue = false;
			Set keySet = conditionalInAssignmentLoop.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "used as condition in loop in which assigned";
		} else if (noDirectUsage && assignCondLoop && !loopUseNoAssign) {
			isFixedValue = false;
			setMessage(11);
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isFixedValue));
		return returnArray;
	}

	/*
	 * Test whether variable is organizer
	 */
	protected ArrayList organizer() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isOrganizer = true;
		if (!isArrayVariable) {
			isOrganizer = false;
			offendingStatement = resultStringer.offendingStatement("organizer");
			reason = "does not appear to be array or is not used directly as array";
		} else if (!isReorganize) {
			isOrganizer = false;
			offendingStatement = resultStringer.offendingStatement("organizer");
			reason = "no organizer type statements found";
		} else if (!onBothSidesOfAssign) {
			isOrganizer = false;
			offendingStatement = resultStringer.offendingStatement("organizer");
			reason = "no statements found where variable on both sides of assignment";
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isOrganizer));
		return returnArray;
	}

	/*
	 * Test whether variable is stepper
	 */
	protected ArrayList stepper() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isStepper = true;
		if (!incDecStatement && !isAssignedInFor && !toggleStatement
				&& !nestedBooleanAssign) {
			isStepper = false;
			offendingStatement = resultStringer.offendingStatement("stepper");
			reason = "no stepper type assign statements found";
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isStepper));
		return returnArray;
	}

	/*
	 * Test whether variable is most recent holder
	 */
	protected ArrayList mostRecentHolder() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isMostRecentHolder = true;
		if (isAssignedInFor) {
			isMostRecentHolder = false;
			setMessage(7);
		} else if (!assignedInLoop) {
			isMostRecentHolder = false;
			setMessage(12);
		} else if (isArrayVariable && isReorganize) {
			isMostRecentHolder = false;
			setMessage(1);
		} else if (isArrayVariable && isFixedValue()) {
			isMostRecentHolder = false;
			Set keySet = assignmentsInLoop.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "is array filled in loop";
		} else if (incDecStatement) {
			isMostRecentHolder = false;
			setMessage(2);
		} else if (toggleStatement) {
			isMostRecentHolder = false;
			setMessage(3);
		} else if (onBothSidesOfAssign) {
			isMostRecentHolder = false;
			setMessage(4);
		} else if (indirectBS) {
			isMostRecentHolder = false;
			setMessage(5);
		} else if (isTransform) {
			isMostRecentHolder = false;
			setMessage(6);
		} else if (conditionForAssignBranch) {
			isMostRecentHolder = false;
			setMessage(9);
		} else if (nestedBooleanAssign) {
			isMostRecentHolder = false;
			setMessage(8);
		} else if (noDirectUsage && assignCondLoop && !loopUseNoAssign) {
			isMostRecentHolder = false;
			setMessage(11);
		} else if (!methodAssignment && !otherAssign) {
			isMostRecentHolder = false;
			setMessage(10);
		} else if (!usedInAssignLoop
				&& (!conditionalUseInAssign || assignmentLoopConditionUse)) {
			isMostRecentHolder = false;
			Set keySet = assignmentsInLoop.keySet();
			Iterator it = keySet.iterator();
			setMessage(13);
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isMostRecentHolder));
		return returnArray;
	}

	/*
	 * Test whether variable is gatherer
	 */
	protected ArrayList gatherer() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isGatherer = true;
		if (isAssignedInFor) {
			isGatherer = false;
			setMessage(7);
		} else if (!assignedInLoop) {
			isGatherer = false;
			setMessage(12);
		} else if (isArrayVariable && isReorganize) {
			isGatherer = false;
			setMessage(1);
		} else if (incDecStatement) {
			isGatherer = false;
			setMessage(2);
		} else if (toggleStatement) {
			isGatherer = false;
			setMessage(3);
		} else if (nestedBooleanAssign) {
			isGatherer = false;
			setMessage(8);
		} else if (conditionForAssignBranch) {
			isGatherer = false;
			setMessage(9);
		} else if (isTransform && !onBothSidesOfAssign && !indirectBS) {
			isGatherer = false;
			setMessage(6);
		} else if (noDirectUsage && assignCondLoop && !loopUseNoAssign
				&& !onBothSidesOfAssign && !indirectBS) {
			isGatherer = false;
			setMessage(11);
		} else if (!methodAssignment && !otherAssign && !onBothSidesOfAssign
				&& !indirectBS) {
			isGatherer = false;
			setMessage(10);
		} else if (!onBothSidesOfAssign && !indirectBS) {
			isGatherer = false;
			Set keySet = assignmentsInLoop.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "not found directly, or indirectly on both sides of assignment statement";
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isGatherer));
		return returnArray;
	}

	/*
	 * Test whether variable is most wanted holder
	 */
	protected ArrayList mostWantedHolder() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isMostWantedHolder = true;
		if (isAssignedInFor) {
			isMostWantedHolder = false;
			setMessage(7);
		} else if (isArrayVariable && isReorganize) {
			isMostWantedHolder = false;
			setMessage(1);
		} else if (incDecStatement) {
			isMostWantedHolder = false;
			setMessage(2);
		} else if (toggleStatement) {
			isMostWantedHolder = false;
			setMessage(3);
		} else if (onBothSidesOfAssign) {
			isMostWantedHolder = false;
			setMessage(4);
		} else if (indirectBS) {
			isMostWantedHolder = false;
			setMessage(5);
		} else if (nestedBooleanAssign) {
			isMostWantedHolder = false;
			setMessage(8);
		} else if (!conditionForAssignBranch) {
			isMostWantedHolder = false;
			HashMap assignments = (HashMap) analysedMap.get("assignment");
			Set keySet = assignments.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "not assigned in branch, or is not condition for branch in which it is assigned";
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isMostWantedHolder));
		return returnArray;
	}

	/*
	 * Test whether variable is one way flag
	 */
	protected ArrayList oneWayFlag() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isOneWayFlag = true;
		if (isAssignedInFor) {
			isOneWayFlag = false;
			setMessage(7);
		} else if (!assignedInLoop) {
			isOneWayFlag = false;
			setMessage(12);
		} else if (isArrayVariable && isReorganize) {
			isOneWayFlag = false;
			setMessage(1);
		} else if (isArrayVariable && !isReorganize) {
			isOneWayFlag = false;
			Set keySet = assignmentsInLoop.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "does not appear to be used as one way flag";
		} else if (incDecStatement) {
			isOneWayFlag = false;
			setMessage(2);
		} else if (toggleStatement) {
			isOneWayFlag = false;
			setMessage(3);
		} else if (onBothSidesOfAssign) {
			isOneWayFlag = false;
			setMessage(4);
		} else if (indirectBS) {
			isOneWayFlag = false;
			setMessage(5);
		} else if (nestedBooleanAssign) {
			isOneWayFlag = false;
			setMessage(8);
		} else if (conditionForAssignBranch) {
			isOneWayFlag = false;
			setMessage(9);
		} else if (isTransform && !assignCondLoop) {
			isOneWayFlag = false;
			setMessage(6);
		} else if (!methodAssignment && !otherAssign) {
			isOneWayFlag = false;
			setMessage(10);
		} else if (!noDirectUsage) {
			isOneWayFlag = false;
			HashMap usages = (HashMap) analysedMap.get("usage");
			Set keySet = usages.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "direct use of variable";
		} else if (!assignCondLoop) {
			isOneWayFlag = false;
			offendingStatement = resultStringer
					.offendingStatement("one way flag");
			reason = "not used directly for assign loop condition";
		} else if (loopUseNoAssign) {
			isOneWayFlag = false;
			offendingStatement = resultStringer
					.offendingStatement("one way flag");
			reason = "used for loop condition outside of loop in which it is assigned";
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isOneWayFlag));
		return returnArray;
	}

	/*
	 * Test whether variable is transformation
	 */
	protected ArrayList transformation() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isTransformation = true;
		if (isAssignedInFor) {
			isTransformation = false;
			setMessage(7);
		} else if (!assignedInLoop) {
			isTransformation = false;
			setMessage(12);
		} else if (isArrayVariable && isReorganize) {
			isTransformation = false;
			setMessage(1);
		} else if (incDecStatement) {
			isTransformation = false;
			setMessage(2);
		} else if (toggleStatement) {
			isTransformation = false;
			setMessage(3);
		} else if (onBothSidesOfAssign) {
			isTransformation = false;
			setMessage(4);
		} else if (indirectBS) {
			isTransformation = false;
			setMessage(5);
		} else if (nestedBooleanAssign) {
			isTransformation = false;
			setMessage(8);
		} else if (conditionForAssignBranch) {
			isTransformation = false;
			setMessage(9);
		} else if (!methodAssignment && !otherAssign && !isTransform) {
			isTransformation = false;
			setMessage(10);
		} else if (methodAssignment) {
			isTransformation = false;
			setMessage(14);
		} else if (!isTransform) {
			isTransformation = false;
			Set keySet = assignmentsInLoop.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "assignment statement contains no operator characters";
		} else if (noDirectUsage && assignCondLoop && !loopUseNoAssign) {
			isTransformation = false;
			setMessage(11);
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isTransformation));
		return returnArray;
	}

	/*
	 * Test whether variable is follower
	 */
	protected ArrayList follower() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isFollower = true;
		if (isAssignedInFor) {
			isFollower = false;
			setMessage(7);
		} else if (!assignedInLoop) {
			isFollower = false;
			setMessage(12);
		} else if (isArrayVariable && isReorganize) {
			isFollower = false;
			setMessage(1);
		} else if (incDecStatement) {
			isFollower = false;
			setMessage(2);
		} else if (toggleStatement) {
			isFollower = false;
			setMessage(3);
		} else if (onBothSidesOfAssign) {
			isFollower = false;
			setMessage(4);
		} else if (indirectBS) {
			isFollower = false;
			setMessage(5);
		} else if (nestedBooleanAssign) {
			isFollower = false;
			setMessage(8);
		} else if (conditionForAssignBranch) {
			isFollower = false;
			setMessage(9);
		} else if (isTransform) {
			isFollower = false;
			setMessage(6);
		} else if (methodAssignment) {
			isFollower = false;
			setMessage(14);
		} else if (noDirectUsage && assignCondLoop && !loopUseNoAssign) {
			isFollower = false;
			setMessage(11);
		} else if (otherAssign) {
			isFollower = false;
			setMessage(15);
		} else if (assignedBeforeUse && !usedOutsideAssign) {
			isFollower = false;
			offendingStatement = (String) assignBeforeUse.get(0);
			reason = "assigned before use in loop and not used outside of assign loop";
		} else if (!usedInAssignLoop) {
			isFollower = false;
			setMessage(13);
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isFollower));
		return returnArray;
	}

	/*
	 * Test whether variable is temporary
	 */
	protected ArrayList temporary() {
		ArrayList returnArray = new ArrayList();
		offendingStatement = "";
		reason = "";
		isTemporary = true;
		if (isAssignedInFor) {
			isTemporary = false;
			setMessage(7);
		} else if (!assignedInLoop) {
			isTemporary = false;
			setMessage(12);
		} else if (isArrayVariable && isReorganize) {
			isTemporary = false;
			setMessage(1);
		} else if (incDecStatement) {
			isTemporary = false;
			setMessage(2);
		} else if (toggleStatement) {
			isTemporary = false;
			setMessage(3);
		} else if (onBothSidesOfAssign) {
			isTemporary = false;
			setMessage(4);
		} else if (indirectBS) {
			isTemporary = false;
			setMessage(5);
		} else if (nestedBooleanAssign) {
			isTemporary = false;
			setMessage(8);
		} else if (conditionForAssignBranch) {
			isTemporary = false;
			setMessage(9);
		} else if (isTransform) {
			isTemporary = false;
			setMessage(6);
		} else if (methodAssignment) {
			isTemporary = false;
			setMessage(14);
		} else if (noDirectUsage && assignCondLoop && !loopUseNoAssign) {
			isTemporary = false;
			setMessage(11);
		} else if (otherAssign) {
			isTemporary = false;
			setMessage(15);
		} else if (!assignedBeforeUse) {
			isTemporary = false;
			Set keySet = usageInAssignmentLoop.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "not assigned in loop before use";
		} else if (usedOutsideAssign) {
			isTemporary = false;
			Set keySet = useOutsideAssign.keySet();
			Iterator it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "used outside of loop in which it is assigned";
		} else if (!usedInAssignLoop) {
			isTemporary = false;
			setMessage(13);
		}
		returnArray.add(offendingStatement);
		returnArray.add(reason);
		returnArray.add(0, new Boolean(isTemporary));
		return returnArray;
	}

	/*
	 * Set message applicable condition detected
	 */
	private void setMessage(int reasonInt) {
		Iterator it;
		Set keySet;
		switch (reasonInt) {
		case 1:
			offendingStatement = (String) reorganize.get(0);
			reason = "appears to be organizer statement";
			break;
		case 2:
			offendingStatement = (String) incDecStatements.get(0);
			reason = "incremented/decremented within loop";
			break;
		case 3:
			offendingStatement = (String) toggleStatements.get(0);
			reason = "toggled within loop";
			break;
		case 4:
			offendingStatement = (String) onBothSides.get(0);
			reason = "appears on both sides of assignment";
			break;
		case 5:
			offendingStatement = (String) indirectBothSides.get(0);
			reason = "indirectly appears on both sides of assignment";
			break;
		case 6:
			offendingStatement = (String) transformations.get(0);
			reason = "assigned in loop with combination of other variables, operators and constants";
			break;
		case 7:
			offendingStatement = (String) assignmentInFor.get(0);
			reason = "assigned in for loop statement";
			break;
		case 8:
			String statementType = (String) nestedBooleanAssignment.get(1);
			if (statementType.equals("nested")) {
				offendingStatement = (String) nestedBooleanAssignment.get(0);
			} else {
				offendingStatement = (String) nestedBooleanAssignment.get(2);
			}
			reason = "appears to be indirectly toggled within loop";
			break;
		case 9:
			keySet = conditionalAssignmentBranch.keySet();
			it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "used as condition for branch in which assigned";
			break;
		case 10:
			keySet = assignmentsInLoop.keySet();
			it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "always assigned in loop with value of other variable";
			break;
		case 11:
			offendingStatement = (String) useForAssignLoopCondition.get(0);
			reason = "condition for loop in which assigned and limited use outside of loop";
			break;
		case 12:
			HashMap assignments = (HashMap) analysedMap.get("assignment");
			keySet = assignments.keySet();
			it = keySet.iterator();
			offendingStatement = (String) it.next();
			reason = "not assigned in loop";
			break;
		case 13:
			HashMap usages = (HashMap) analysedMap.get("usage");
			keySet = usages.keySet();
			if (!keySet.isEmpty()) {
				it = keySet.iterator();
				offendingStatement = (String) it.next();
			} else {
				HashMap conditionals = (HashMap) analysedMap.get("conditional");
				keySet = conditionals.keySet();
				it = keySet.iterator();
				offendingStatement = (String) it.next();
			}
			reason = "not used in loop in which assigned";
			break;
		case 14:
			offendingStatement = (String) methodAssignments.get(0);
			reason = "assigned with output from call to method";
			break;
		case 15:
			offendingStatement = (String) otherAssignments.get(0);
			reason = "assigned directly with value, instantiation of object, or call to method of other object";
			break;
		}
	}
}